/*
** Module   : ComboBoxModel.java
** Abstract : Combo-box data model.
**
** Copyright (c) 2010-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20100629          Created initial version
** 002 SIY 20100804          Changes in imports, some refactorings.
** 003 IAS 20150202          Add selected index.
*/

package com.goldencode.p2j.ui.client.model;

/**
 * Combo-box data model
 * 
 * @param   <T>
 *          Type of the item stored in the model. 
 */
public class ComboBoxModel<T>
extends ListModel<T>
{
   /** Selected item. */
   private T selected;
   /** Selected index. */
   private int selectedIndex;

   /**
    * Add new item at the end of model.
    * 
    * @param   item
    *          Item to add.
    */
   @Override
   public void add(T item)
   {
      super.add(item);

      if (size() == 1)
         selected = item;
   }

   /**
    * Insert item at specified location.
    *  
    * @param   item
    *          Item to insert.
    * @param   index
    *          Index at which item should be inserted.
    */
   public void insert(T item, int index)
   {
      add(index, item);

      if (size() == 1)
         selected = item;
   }

   /**
    * Remove item with specified value.
    * 
    * @param   item
    *          Item to remove.
    */
   public void remove(T item)
   {
      int index = indexOf(item);
      
      if (index >= 0)
         removeAt(index);
   }

   /**
    * Remove all items.
    */
   @Override
   public void clear()
   {
      super.clear();
      selected = null;
   }

   /**
    * Remove item at specified position.
    * 
    * @param   index
    *          Item index.
    */
   @Override
   public void removeAt(int index)
   {
      if (selected == get(index))
      {
         if (index == 0)
            selected = (size() == 1) ? null : get(1);
         else
            selected = get(index - 1);
      }

      super.removeAt(index);
   }

   /**
    * Get selected item.
    * 
    * @return  selected item.
    */
   public T selected()
   {
      return selected;
   }

   /**
    * Get selected index.
    * 
    * @return  selected index.
    */
   public int selectedIndex()
   {
      return selectedIndex;
   }

   /**
    * Select item with specified value.
    * 
    * @param   item
    *          Item value to select.
    */
   public void select(T item)
   {
      selected = item;
   }

   /**
    * Set selected index.
    * 
    * @param   index
    *          Selected item index.
    */
   public void index(int index)
   {
      selectedIndex = index;
   }
}
